import numpy as np
import pandas as pd


def check_new_observation(new_observation, explainer):
    if isinstance(new_observation, (pd.Series, list)):
        new_observation = np.array(new_observation)
        if new_observation.ndim == 1:
            # make 1D array 2D
            new_observation = new_observation.reshape((1, -1))

        new_observation = pd.DataFrame(new_observation, columns=explainer.data.columns)
    elif isinstance(new_observation, pd.DataFrame):
        new_observation.columns = explainer.data.columns
    else:
        raise TypeError("new_observation must be a numpy.ndarray or pandas.Series or pandas.DataFrame")

    return new_observation


def check_order(order):
    if order is not None and not isinstance(order, (list, np.ndarray)):
        raise TypeError('Wrong order type')

    if order is not None:
        return np.array(order)
    else:
        return None



import numpy as np
import pandas as pd


def check_variables(variables, explainer):
    if variables is not None and isinstance(variables, (list, np.ndarray, pd.Series)):
        if not set(variables).issubset(explainer.data.columns):
            raise ValueError('Invalid variable names')
    elif variables is not None and not isinstance(variables, (list, np.ndarray, pd.Series)):
        raise TypeError("Variables must by list or numpy.ndarray or pd.Series")
    else:
        variables = explainer.data.columns

    return list(variables)


def check_new_observation(new_observation, explainer):
    if not isinstance(new_observation, (pd.DataFrame,)):
        raise TypeError("new_observation must be the pandas.DataFrame")

    if new_observation.index.unique().shape[0] != new_observation.shape[0]:
        raise ValueError("Index is not unique")

    return new_observation.copy()


def check_data(data, variables):
    if data is None:
        raise ValueError("Data must be provided by explainer")

    return data


def check_y(y):
    if y is not None and isinstance(y, pd.Series):
        y = np.array(y)
    elif y is not None and not isinstance(y, np.ndarray):
        raise TypeError("y must be numpy.ndarray or pandas.Series")

    return y


def check_variable_splits(variable_splits, variables):
    """
    Validate variable splits
    """
    if variable_splits is not None and set(variable_splits.keys()) == set(variables):
        return True
    else:
        return False



def yhat_default(m, d):
    return m.predict(d)


def yhat_proba(m, d):
    return m.predict_proba(d)[:, 1]


def yhat(model):
    if hasattr(model, 'predict_proba'):
        # check if model has predict_proba
        return yhat_proba
    elif hasattr(model, 'predict'):
        # check if model has predict
        return yhat_default

    return False



import numpy as np


def check_variable_type(variable_type):
    if variable_type not in ['numerical', 'categorical']:
        raise ValueError("variable_type needs to be 'numerical' or 'categorical'")


def prepare_all_variables(all_profiles, variables):
    # variables to use
    all_variables = all_profiles['_vname_'].dropna().unique().astype(str)

    if variables is not None:
        all_variables_intersect = set(all_variables).intersection(set(variables))
        if len(all_variables_intersect) == 0:
            raise ValueError("variables do not overlap with " + all_variables)
        all_variables = np.array(list(all_variables_intersect))

    return all_variables


def prepare_numerical_categorical(all_variables, all_profiles, variable_type):
    # only numerical or only factors?
    is_numeric = np.empty_like(all_variables, bool)
    for i, var in enumerate(all_variables):
        is_numeric[i] = np.issubdtype(all_profiles[var].dtype, np.number)

    if variable_type == 'numerical':
        vnames = all_variables[is_numeric]
        if vnames.shape[0] == 0:
            raise ValueError("There are no numerical variables")

        all_profiles['_x_'] = 0

    else:
        vnames = all_variables[~is_numeric]
        if vnames.shape[0] == 0:
            raise ValueError("There are no non-numerical variables")

        all_profiles['_x_'] = ""

    return all_profiles, vnames


def create_x(all_profiles, variable_type):
    # create _x_
    for variable in all_profiles['_vname_'].unique():
        where_variable = all_profiles['_vname_'] == variable
        all_profiles.loc[where_variable, '_x_'] = all_profiles.loc[where_variable, variable]

    # change x column to proper character values
    if variable_type == 'categorical':
        all_profiles.loc[:, '_x_'] = all_profiles.apply(lambda row: str(row[row['_vname_']]), axis=1)

    return all_profiles


def check_groups(groups):
    if groups is not None and not isinstance(groups, str):
        raise TypeError("groups must be None or str")

    # treating groups as list simplifies code
    return [] if groups is None else [groups]



import numpy as np
import pandas as pd


def calculate_variable_importance(explainer,
                                  type,
                                  loss_function,
                                  variables,
                                  N,
                                  B,
                                  label,
                                  keep_raw_permutations):
    result = [None] * B

    for i in range(B):
        result[i] = loss_after_permutation(explainer, loss_function, variables, N)

    raw = pd.concat(result, sort=True)
    result = raw.mean().sort_values().reset_index()
    result['label'] = label

    result.rename(columns={0: 'dropout_loss', 'index': 'variable'}, inplace=True)

    if type == "ratio":
        result.loc[:, 'dropout_loss'] = result.loc[:, 'dropout_loss'] / result.loc[
            result.variable == '_full_model_', 'dropout_loss'].values

    if type == "difference":
        result.loc[:, 'dropout_loss'] = result.loc[:, 'dropout_loss'] - result.loc[
            result.variable == '_full_model_', 'dropout_loss'].values

    raw_permutations = raw.reset_index(drop=True) if keep_raw_permutations else None

    return result, raw_permutations


def loss_after_permutation(explainer, loss_function, variables, N):
    if N is not None:
        sampled_rows = np.random.choice(range(explainer.data.shape[0]), N, False)
    else:
        sampled_rows = np.arange(explainer.data.shape[0])

    sampled_data = explainer.data.iloc[sampled_rows, :]

    observed = explainer.y[sampled_rows]

    # loss on the full model or when outcomes are permuted
    loss_full = loss_function(observed, explainer.predict(sampled_data))

    sampled_rows2 = np.random.choice(range(observed.shape[0]), observed.shape[0], False)
    loss_baseline = loss_function(observed[sampled_rows2], explainer.predict(sampled_data))

    loss_features = {}
    for variables_set_key in variables:
        ndf = sampled_data.copy()
        ndf.loc[:, variables[variables_set_key]] = ndf.iloc[
            np.random.choice(range(ndf.shape[0]), ndf.shape[0], False), :].loc[:, variables[variables_set_key]].values

        predicted = explainer.predict(ndf)

        loss_features[variables_set_key] = loss_function(observed, predicted)

    loss_features['_full_model_'] = loss_full
    loss_features['_baseline_'] = loss_baseline

    return pd.DataFrame(loss_features, index=[0])



from ._model_performance.object import ModelPerformance
from ._variable_importance.object import VariableImportance
from ._aggregated_profiles.object import AggregatedProfiles

__all__ = [
    "ModelPerformance",
    "VariableImportance",
    "AggregatedProfiles"
]



import pandas as pd
import os


def load_titanic():
    abs_dir_path = os.path.dirname(os.path.abspath(__file__))
    abs_datasets_path = os.path.join(abs_dir_path, 'data', 'titanic.csv')

    dataset = pd.read_csv(abs_datasets_path)

    return dataset



def tooltip_text(obs, r=None):
    temp = ""
    if r is not None:
        for var in obs.index:
            if var == "_yhat_":
                temp += "prediction:<br>" + "- before: " + str(obs[var]) + "<br>" + "- after: " +\
                        str(r[var]) + "<br><br>"
            elif var == r['_vname_']:
                temp += str(var) + ": " + str(r['_xhat_']) + "</br>"
            else:
                temp += str(var) + ": " + str(obs[var]) + "</br>"
    else:
        for var in obs.index:
            if var == "_yhat_":
                temp += "prediction:" + str(obs[var]) + "<br><br>"
            else:
                temp += str(var) + ": " + str(obs[var]) + "</br>"
    return temp



from dalex.instance_level import BreakDown, Shap, CeterisParibus
from dalex.dataset_level import ModelPerformance, VariableImportance, AggregatedProfiles
from .checks import *


class Explainer:
    def __init__(self,
                 model,
                 data=None,
                 y=None,
                 predict_function=None,
                 residual_function=None,
                 weights=None,
                 label=None,
                 verbose=True,
                 precalculate=True,
                 colorize=True,
                 model_info=None,
                 model_type=None):

        """Create Model Explainer

        Black-box models may have very different structures.
        This function creates a unified representation of a model, which can be further processed by various explainers.
        Please NOTE, that the model is the only required argument.
        But some explainers may require that other arguments will be provided too.

        :param model: a model to be explained
        :param data: data.frame or matrix - data that was used for fitting. If not provided then will be extracted from the model. Data should be passed without target column (this shall be provided as the y argument). NOTE: If target variable is present in the data, some of the functionalities my not work properly.
        :param y: numeric vector with outputs / scores. If provided then it shall have the same size as data
        :param predict_function: function that takes two arguments: model and new data and returns numeric vector with predictions
        :param residual_function: function that takes three arguments: model, data and response vector y. It should return a numeric vector with model residuals for given data. If not provided, response residuals y-yhat are calculated.
        :param weights: weights numeric vector with sampling weights. By default it's None. If provided then it shall have the same length as data
        :param label: the name of the model. By default it's extracted from the 'class' attribute of the model
        :param verbose: if True (default) then diagnostic messages will be printed
        :param precalculate: if True (default) then predicted_values and residual are calculated when explainer is created.
        :param colorize: TODO
        :param model_info: list containg information about the model
        :param model_type: type of a model, either classification or regression.
        """

        verbose_cat("Preparation of a new explainer is initiated\n", verbose=verbose)

        # if requested, remove colors
        if not colorize:
            color_codes = {
                'yellow_start': "",
                'yellow_end': "",
                'red_start': "",
                'red_end': "",
                'green_start': "",
                'green_end': ""}

        # REPORT: checks for model label
        label = check_label(label, model, verbose)

        # REPORT: checks for data
        """
        Contrary to DALEX in R, data cannot be retrieved from model, thus data is necessary.
        if data is None
        tell user about the lack of the data
        """

        """
        Input numpy/pandas
        """

        data = check_data(data, verbose)

        # REPORT: checks for y
        y = check_y(y, data, verbose)

        # REPORT: checks for weights
        weights = check_weights(weights, data, verbose)

        # REPORT: checks for predict_function
        predict_function, pred = check_predict_function(predict_function, model, data, precalculate, verbose)

        # if data is specified then we may test predict_function
        # at this moment we have predict function

        # REPORT: checks for residual_function
        residual_function, residuals = check_residual_function(residual_function, predict_function, model, data, y,
                                                               precalculate, verbose)

        model_info = check_model_info(model_info, model, verbose)

        # READY to create an explainer
        self.model = model
        self.data = data
        self.y = y
        self.predict_function = predict_function
        self.y_hat = pred
        self.residual_function = residual_function
        self.residuals = residuals
        self.model_class = type(model)
        self.label = label
        self.model_info = model_info
        self.weights = weights
        self.model_type = model_type

        verbose_cat("\nA new explainer has been created!", verbose=verbose)

    def predict(self, data):

        """This is a generic function for making a prediction.
        
        :param data: pd.DataFrame
        :return: array-like, prediction of the model
        """

        return self.predict_function(self.model, data)

    def residual(self, data, y):

        """This is a generic function for calculating residuals.

        :param data: pd.DataFrame
        :param y: numeric vector with outputs / scores. If provided then it shall have the same size as data
        :return: array-like, prediction of the model
        """

        return self.residual_function(self.model, data, y)

    def predict_parts(self,
                     new_observation,
                     type=('break_down_interactions','break_down','shap'),
                     order=None,
                     interaction_preference=1,
                     path="average",
                     B=25,
                     keep_distributions=False):

        """Instance Level Variable Attribution as Break Down or SHAP Explanations

        :param new_observation: pd.Series, a new observarvation for which predictions need to be explained
        :param type: type the type of variable attributions. Either 'shap', 'break_down' or 'break_down_interactions'
        :param order: if not `NULL`, then it will be a fixed order of variables. It can be a numeric vector or vector with names of variables.
        :param interaction_preference: an integer specifying which interactions will be present in an explanation. The larger the integer, the more frequently interactions will be presented.
        :param path: if specified, then this path will be highlighted on the plot. Use `average` in order to show an average effect
        :param B: number of random paths
        :param keep_distributions: if `TRUE`, then distribution of partial predictions is stored and can be plotted with the generic `plot()`.
        :return: BreakDown / Shap
        """

        types = ('break_down_interactions','break_down','shap')
        type = check_method_type(type, types)

        if type == 'break_down_interactions' or type == 'break_down':
            predict_parts_ = BreakDown(
                type=type,
                keep_distributions=keep_distributions,
                order=order,
                interaction_preference=interaction_preference
            )
        elif type == 'shap':
            predict_parts_ = Shap(
                keep_distributions=keep_distributions,
                path=path,
                B=B
            )

        predict_parts_.fit(self, new_observation)

        return predict_parts_

    def predict_profile(self,
                        new_observation,
                        type=('ceteris_paribus'),
                        y=None,
                        variables=None,
                        grid_points=101,
                        variable_splits=None):

        """Creates CeterisParibus object

        :param new_observation: DataFrame with observations for which the profiles will be calculated
        :param type TODO
        :param y: pandas Series with labels for the observations
        :param variables: variables for which the profiles are calculated
        :param grid_points: number of points in a single variable split if calculated automatically
        :param variable_splits: mapping of variables into points the profile will be calculated, if None then calculate with the function `_calculate_variable_splits`
        :return CeterisParibus object
        """

        types = ('ceteris_paribus')
        type = check_method_type(type, types)

        if type == 'ceteris_paribus':
            predict_profile_ = CeterisParibus(
                variables=variables,
                grid_points=grid_points,
                variable_splits=variable_splits
            )

        predict_profile_.fit(self, new_observation, y)

        return predict_profile_

    def model_performance(self,
                          model_type=None,
                          cutoff=0.5):

        """Model Performance Measures

        :param model_type: type of prediction regression/classification
        :param cutoff: a cutoff for classification models, needed for measures like recall, precision, ACC, F1. By default 0.5.
        :return: ModelPerformance object
        """

        if model_type is None and self.model_type is None:
            raise TypeError("if self.model_type is None, then model_type must be not None")
        elif model_type is None:
            model_type = self.model_type

        model_performance_ = ModelPerformance(
            model_type=model_type,
            cutoff=cutoff
        )
        model_performance_.fit(self)

        return model_performance_

    def model_parts(self,
                    loss_function='loss_root_mean_square',
                    type=('variable_importance','ratio','difference'),
                    N=None,
                    B=10,
                    keep_raw_permutations=None,
                    variables=None,
                    variable_groups=None,
                    random_state=None):

        """Creates VariableImportance object

        :param loss_function: a function thet will be used to assess variable importance
        :param type: type of transformation that should be applied for dropout loss
        :param N: number of observations that should be sampled for calculation of variable importance
        :param B: number of permutation rounds to perform on each variable
        :param keep_raw_permutations: TODO
        :param variables: vector of variables. If None then variable importance will be tested for each variable from the data separately
        :param variable_groups: list of variables names vectors. This is for testing joint variable importance
        :param label: TODO
        :param random_state: random state for the permutations
        :return: FeatureImportance object
        """

        types = ('variable_importance','ratio','difference')
        type = check_method_type(type, types)

        model_parts_ = VariableImportance(
            loss_function=loss_function,
            type=type,
            N=N,
            B=B,
            variables=variables,
            variable_groups=variable_groups,
            random_state=random_state,
            keep_raw_permutations=keep_raw_permutations,
        )

        model_parts_.fit(self)

        return model_parts_

    def model_profile(self,
                      type=('partial','accumulated','conditional'),
                      N=500,
                      variables=None,
                      variable_type='numerical',
                      groups=None,
                      span=0.25,
                      grid_points=101,
                      intercept=True):

        """Dataset Level Variable Effect as Partial Dependency Profile or Accumulated Local Effects

        :param ceteris_paribus: a ceteris paribus explainer or list of ceteris paribus explainers
        :param N: number of observations used for calculation of partial dependency profiles. By default, 500 observations will be chosen randomly.
        :param variables: names of variables for which profiles shall be calculated.
        :param variable_type: TODO If "numerical" then only numerical variables will be calculated. If "categorical" then only categorical variables will be calculated.
        :param groups: a variable name that will be used for grouping.
        :param type: either partial/conditional/accumulated for partial dependence, conditional profiles of accumulated local effects
        :param span: smoothing coeffcient, by default 0.25.It's the sd for gaussian kernel
        :param grid_points: number of points for profile
        :param intercept: False if center data on 0
        :return: VariableEffect object
        """

        types = ('partial','accumulated','conditional')
        type = check_method_type(type, types)

        N = min(N, self.data.shape[0])
        I = np.random.choice(np.arange(N), N, replace=False)

        ceteris_paribus = CeterisParibus(grid_points=grid_points)
        ceteris_paribus.fit(self, self.data.iloc[I, :], self.y[I])

        model_profile_ = AggregatedProfiles(
            type=type,
            variables=variables,
            variable_type=variable_type,
            groups=groups,
            span=span,
            intercept=intercept
        )

        model_profile_.fit(ceteris_paribus)

        return model_profile_

    def dumps(self, *args, **kwargs):
        """Return binary (pickled) representation of the explainer.

        Note, that local functions and lambdas cannot be pickled.
        Residual function by default contains lambda, thus, if default, is always dropped.
        Original explainer is not changed.

        :param args: Positional arguments passed to pickle.dumps
        :param kwargs: Keyword arguments passed to pickle.dumps
        :return: binary (pickled) representation of the explainer
        """

        from copy import deepcopy
        to_dump = deepcopy(self)
        to_dump = check_if_local_and_lambda(to_dump)

        import pickle
        return pickle.dumps(to_dump, *args, **kwargs)

    def dump(self, file, *args, **kwargs):
        """Save binary (pickled) representation of the explainer to the file.

        Note, that local functions and lambdas cannot be pickled.
        Residual function by default contains lambda, thus, if default, is always dropped.
        Original explainer is not changed.

        :param file: It can be a file object opened for binary writing, an io.BytesIO instance.
        :param args: Positional arguments passed to pickle.dump
        :param kwargs: Keyword arguments passed to pickle.dump
        :return: binary (pickled) representation of the explainer
        """

        from copy import deepcopy
        to_dump = deepcopy(self)
        to_dump = check_if_local_and_lambda(to_dump)

        import pickle
        return pickle.dump(to_dump, file, *args, **kwargs)

    @staticmethod
    def loads(data, use_defaults=True, *args, **kwargs):
        """Load explainer from binary (pickled) representation.

        Note, that local functions and lambdas cannot be pickled.
        If use_defaults is set to True, then dropped functions are set to defaults.

        :param data: binary representation of the explainer
        :param use_defaults: True, if dropped functions should be replaced with defaults
        :param args: Positional arguments passed to pickle.loads
        :param kwargs: Keyword arguments passed to pickle.loads
        :return: Explainer
        """

        import pickle
        exp = pickle.loads(data, *args, **kwargs)

        if use_defaults:
            exp = check_if_empty_fields(exp)

        return exp

    @staticmethod
    def load(file, use_defaults=True, *args, **kwargs):
        """Load explainer from binary file (pickled).

        Note, that local functions and lambdas cannot be pickled.
        If use_defaults is set to True, then dropped functions are set to defaults.

        :param file: It can be a binary file object opened for reading, an io.BytesIO object
        :param use_defaults: True, if dropped functions should be replaced with defaults
        :param args: Positional arguments passed to pickle.load
        :param kwargs: Keyword arguments passed to pickle.load
        :return: Explainer
        """

        import pickle
        exp = pickle.load(file, *args, **kwargs)

        if use_defaults:
            exp = check_if_empty_fields(exp)

        return exp



import unittest
import dalex as dx

import pandas as pd
import numpy as np

from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.preprocessing import StandardScaler, LabelEncoder, OneHotEncoder
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.tree import DecisionTreeRegressor
from sklearn.compose import ColumnTransformer


class PredictTestTitanic(unittest.TestCase):
    def setUp(self):
        data = pd.read_csv("titanic.csv", index_col=0).dropna()
        data.loc[:, 'survived'] = LabelEncoder().fit_transform(data.survived)

        self.X = data.drop(columns='survived')
        self.y = data.survived

        numeric_features = ['age', 'fare', 'sibsp', 'parch']
        numeric_transformer = Pipeline(steps=[
            ('imputer', SimpleImputer(strategy='median')),
            ('scaler', StandardScaler())])

        categorical_features = ['gender', 'class', 'embarked', 'country']
        categorical_transformer = Pipeline(steps=[
            ('imputer', SimpleImputer(strategy='constant', fill_value='missing')),
            ('onehot', OneHotEncoder(handle_unknown='ignore'))])

        preprocessor = ColumnTransformer(
            transformers=[
                ('num', numeric_transformer, numeric_features),
                ('cat', categorical_transformer, categorical_features)])

        clf = Pipeline(steps=[('preprocessor', preprocessor),
                              ('classifier', MLPRegressor(hidden_layer_sizes=(150, 100, 50),
                                                          max_iter=500, random_state=0))])

        clf.fit(self.X, self.y)

        self.exp = dx.Explainer(clf, self.X, self.y)

    def test_result(self):
        johny_d = pd.DataFrame({'gender': ['male'],
                                'age': [8],
                                'class': ['1st'],
                                'embarked': ['Southampton'],
                                'country': ['England'],
                                'fare': [72],
                                'sibsp': [0],
                                'parch': 0})

        henry = pd.DataFrame({'gender': ['male'],
                              'age': [47],
                              'class': ['1st'],
                              'embarked': ['Cherbourg'],
                              'country': ['United States'],
                              'fare': [25],
                              'sibsp': [0],
                              'parch': [0]})

        self.assertAlmostEqual(self.exp.predict(johny_d)[0], 0.7564213)
        self.assertAlmostEqual(self.exp.predict(henry)[0], 0.15350221)


if __name__ == '__main__':
    unittest.main()


from .loss_functions import *


def check_loss_function(loss_function):
    if isinstance(loss_function, str):
        if loss_function == 'loss_root_mean_square':
            loss_function = loss_root_mean_square

    return loss_function


def check_variable_groups(variable_groups):
    if variable_groups is not None:
        if not isinstance(variable_groups, dict):
            raise TypeError("variable_groups should be of class dict")

        wrong_names = np.empty(len(variable_groups))
        for i, key in enumerate(variable_groups):
            if isinstance(variable_groups[key], list):
                variable_groups[key] = np.array(variable_groups[key])
            elif not isinstance(variable_groups[key], np.ndarray):
                raise TypeError("variable_groups' elements must be list of strings or numpy.ndarray with strings")

            if not isinstance(variable_groups[key][0], str):
                raise TypeError("variable_groups' elements must be list of strings or numpy.ndarray with strings")

            wrong_names[i] = np.in1d(variable_groups[key], explainer.data.columns)

        wrong_names = not wrong_names.all()

        if wrong_names:
            raise ValueError("You have passed wrong variables names in variable_groups argument")

    return variable_groups


def check_B(B):
    return max(1, np.round(B))


def check_variables(variables, variable_groups, explainer):
    if variable_groups is None:
        # if `variables` are not specified, then extract from data
        if variables is None:
            variables = list(explainer.data.columns)
    else:
        variables = variable_groups

    if not isinstance(variables, (list, np.ndarray, dict)):
        raise TypeError("variables must be list or numpy.ndarray or a dict")

    if isinstance(variables, (list, np.ndarray)):
        result = {}
        for var in variables:
            result[var] = var

        variables = result

    return variables


def check_label(label, explainer):
    if label is None:
        label = explainer.label

    return label


def check_type(type):
    if not isinstance(type, (str, tuple, list)):
        raise TypeError("type must be a string or tuple or list")

    if isinstance(type, (tuple, list)):
        type = type[0]

    return type


def check_random_state(random_state):
    if random_state is not None:
        np.random.set_state(random_state)

    return random_state


def check_keep_raw_permutations(keep_raw_permutations, B):
    return B > 1 if keep_raw_permutations is None else keep_raw_permutations



import setuptools
from os import path


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="dalex",
    version="0.1.4",
    author="Wojciech Kretowicz, Hubert Baniecki, Przemyslaw Biecek",
    author_email="wojtekkretowicz@gmail.com, hbaniecki@gmail.com",
    description="DALEX in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ModelOriented/DALEX",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'pandas>=1.0.1',
        'numpy>=1.18.1',
        'plotly>=4.5.2',
        'tqdm>=4.42.1'
    ],
    packages=setuptools.find_packages(include=["dalex", "dalex.*"]),
    python_requires='>=3.6',
    include_package_data=True
)



import numpy as np
import pandas as pd
from tqdm import tqdm


def calculate_ceteris_paribus(explainer,
                              new_observation,
                              variable_splits,
                              y):
    """
    Inner method that calculates ceteris paribus and some additional fields.

    Set profile and new_profile fields to actual values.

    :return: None
    """
    
    profiles = calculate_variable_profile(explainer,
                                          new_observation,
                                          variable_splits)

    profiles.loc[:, '_label_'] = explainer.label

    # add points of interests
    predictions = explainer.predict(new_observation)

    new_observation.loc[:, '_yhat_'] = predictions

    new_observation.loc[:, '_label_'] = explainer.label

    new_observation.loc[:, '_ids_'] = new_observation.index.values

    if y is not None:
        new_observation.loc[:, '_y_'] = y

    return profiles, new_observation


def calculate_variable_profile(explainer,
                               data,
                               variable_splits):
    """
    Inner function that calculates ceteris paribus for all variables.

    :param explainer:
    :param data: new observations
    :param variable_splits: dictionary of split points
    :return: ceteris paribus profile for all variable
    """

    profile = []
    for variable in tqdm(variable_splits, desc="Calculating ceteris paribus!"):
        split_points = variable_splits[variable]
        profile.append(single_variable_profile(explainer, data, variable, split_points))

    return pd.concat(profile)


def single_variable_profile(explainer,
                            data,
                            variable,
                            split_points):
    """
    Inner function for calculating variable profile. This function is iterated over all variables.

    :param explainer:
    :param data: new observations
    :param variable: considered variable
    :param split_points: dataset is split over these points
    :return: ceteris paribus profile for one variable
    """
    # remember ids of selected points
    ids = np.repeat(data.index, split_points.shape[0])
    new_data = data.loc[ids, :]
    new_data.loc[:, variable] = np.tile(split_points, data.shape[0])

    yhat = explainer.predict(new_data)

    new_data.loc[:, '_yhat_'] = yhat
    new_data.loc[:, '_vname_'] = variable
    new_data.loc[:, '_ids_'] = ids

    return new_data


def calculate_variable_split(data, variables, grid_points):
    """
    Calculate points for splitting the dataset

    :param data: dataset to split
    :param variables: variables to calculate ceteris paribus
    :param grid_points: how many points should split the dataset
    :return: dict, dictionary of split points for all variables
    """
    variable_splits = {}
    for variable in variables:
        if np.issubdtype(data.loc[:, variable].dtype, np.floating):
            # grid points might be larger than the number of unique values
            probs = np.linspace(0, 1, grid_points)

            variable_splits[variable] = np.unique(np.quantile(data.loc[:, variable], probs))
        else:
            variable_splits[variable] = np.unique(data.loc[:, variable])

    return variable_splits
